import {
  Badge,
  Box,
  Button,
  Grid,
  IconButton,
  Paper,
  SwipeableDrawer,
  Typography,
  useTheme,
} from "@mui/material";
import createStyles from "@mui/styles/createStyles";
import makeStyles from "@mui/styles/makeStyles";
import { Field, Form, Formik, FormikHelpers, FormikProps } from "formik";
import { FC, useCallback, useRef, useState } from "react";
import { Editor } from "../commons/editor/vditor";
import * as Yup from "yup";
import {
  Article,
  CreateArticleInput,
  ArticleHistory,
  UpdateArticleInput,
} from "../generated/graphql";
import { DateTimePicker } from "formik-material-ui-pickers";
import { gql, useMutation } from "@apollo/client";
import * as R from "ramda";
import { useNavigate } from "react-router-dom";
import { useDrafts } from "./hooks/useDrafts";
import Timeline from "@mui/lab/Timeline";
import TimelineItem from "@mui/lab/TimelineItem";
import TimelineSeparator from "@mui/lab/TimelineSeparator";
import TimelineConnector from "@mui/lab/TimelineConnector";
import TimelineContent from "@mui/lab/TimelineContent";
import TimelineDot from "@mui/lab/TimelineDot";
import { HistoryEduOutlined } from "@mui/icons-material";
import { format } from "date-fns";

const CREATE_ARTICLE = gql`
  mutation createArticle($createArticleInput: CreateArticleInput!) {
    createArticle(createArticleInput: $createArticleInput) {
      id
      title
      content
      publishedAt
    }
  }
`;

const UPDATE_ARTICLE = gql`
  mutation updateArticle($updateArticleInput: UpdateArticleInput!) {
    updateArticle(updateArticleInput: $updateArticleInput) {
      id
      title
      content
      publishedAt
    }
  }
`;

const useStyles = makeStyles((theme) =>
  createStyles({
    form: {
      padding: "15px 30px",
    },
    editor: {
      height: "700px",
    },
  })
);

const fields = ["title", "content", "publishedAt", "tags"] as const;
type Values = Pick<Article, typeof fields[number]> &
  Partial<Pick<Article, "id">>;
interface Props {
  article?: Values;
  draftKey: string;
}

export const ArticleEditor: FC<Props> = ({ article, draftKey }) => {
  const classes = useStyles();
  const navigate = useNavigate();
  const formik = useRef<FormikProps<Values>>(null);
  const { draft, putDraft, drafts, getDraft } = useDrafts(draftKey);

  const validationSchema = Yup.object({
    content: Yup.string()
      .max(65535, "文章内容不得超过 65535 个字符")
      .required("文章内容不得为空"),
    publishedAt: Yup.date(),
  });

  const [createArticle] = useMutation<
    { createArticle: Article },
    {
      createArticleInput: CreateArticleInput;
    }
  >(CREATE_ARTICLE, {
    update(cache, { data }) {
      cache.modify({
        fields: {
          articles(existingArticles = []) {
            const newArticleRef = cache.writeFragment({
              data: data!.createArticle,
              fragment: gql`
                fragment NewArticle on Article {
                  id
                  title
                  content
                  publishedAt
                }
              `,
            });
            return [newArticleRef, ...existingArticles];
          },
        },
      });
    },
  });
  const [updateArticle] = useMutation<{
    updateArticleInput: UpdateArticleInput;
  }>(UPDATE_ARTICLE);

  const submitForm = async (
    values: Values,
    { setSubmitting }: FormikHelpers<Values>
  ) => {
    if ("id" in article!) {
      const title = /# (.+)$/m.exec(values.content ?? "")?.[1];
      if (!title) {
        console.log("no title");
        setSubmitting(false);
        return;
      }
      await updateArticle({
        variables: {
          updateArticleInput: {
            ...values,
            title,
          },
        },
      })
        .then(() => {
          navigate("/articles", { replace: true });
        })
        .finally(() => {
          setSubmitting(false);
        });
    } else {
      const title = /# (.+)$/m.exec(values.content ?? "")?.[1];
      if (!title) {
        console.log("no title");
        setSubmitting(false);
        return;
      }
      await createArticle({
        variables: {
          createArticleInput: {
            ...(values as CreateArticleInput),
            title,
          },
        },
      })
        .then(() => {
          navigate("/articles", { replace: true });
        })
        .finally(() => {
          setSubmitting(false);
        });
    }
    setSubmitting(false);
  };

  const handleSave = useCallback(
    async (automatically = true) => {
      await putDraft({
        key: draftKey,
        payload: formik.current?.values,
        automatically,
      });
    },
    [draftKey, putDraft, formik]
  );

  return (
    <Paper>
      <Formik
        initialValues={{
          ...R.pick(fields, article!),
        }}
        innerRef={formik}
        validationSchema={validationSchema}
        autoComplete="off"
        onSubmit={submitForm}
      >
        {({ submitForm, isSubmitting }) => (
          <Form className={classes.form}>
            <Grid container spacing={3}>
              <Grid item xs={12}>
                <Editor
                  className={classes.editor}
                  name="content"
                  label="Content"
                />
              </Grid>
              <Grid item xs={12}>
                <Field
                  ampm={false}
                  component={DateTimePicker}
                  label="Published At"
                  name="publishedAt"
                  format="yyyy-MM-dd HH:mm:ss"
                />
              </Grid>
              <Grid item xs={12}>
                <Button
                  variant="contained"
                  color="primary"
                  disabled={isSubmitting}
                  onClick={() => handleSave(false)}
                >
                  Save
                </Button>
                <Button
                  style={{ marginLeft: "10px" }}
                  variant="contained"
                  color="secondary"
                  disabled={isSubmitting}
                  onClick={submitForm}
                >
                  Publish
                </Button>

                <HistoryDrawer
                  drafts={drafts}
                  setCurrentDraft={(draft) => getDraft(draft.id)}
                />
              </Grid>
            </Grid>
          </Form>
        )}
      </Formik>
    </Paper>
  );
};

interface HistoryDrawerProps {
  drafts?: DraftArticle[];
  setCurrentDraft: (draft: DraftArticle) => void;
}

const HistoryDrawer: FC<HistoryDrawerProps> = ({ drafts, setCurrentDraft }) => {
  const theme = useTheme();
  const [open, setOpen] = useState(false);
  return (
    <>
      <IconButton aria-label="history" onClick={() => setOpen(!open)}>
        <Badge color="secondary" badgeContent={drafts?.length}>
          <HistoryEduOutlined />
        </Badge>
      </IconButton>
      <SwipeableDrawer
        container={window.document.body}
        anchor="right"
        open={open}
        onClose={() => setOpen(false)}
        onOpen={() => setOpen(true)}
        swipeAreaWidth={50}
        disableSwipeToOpen={false}
        ModalProps={{
          keepMounted: true,
          BackdropProps: {
            sx: {
              backgroundColor: "transparent",
            },
          },
        }}
      >
        <Box
          sx={{
            borderTopLeftRadius: 8,
            borderTopRightRadius: 8,
            visibility: "visible",
            height: "100%",
            width: "200px",
            overflowY: "auto",
            overflowX: "hidden",
            mt: theme.mixins.toolbar.minHeight,
          }}
        >
          <Timeline>
            {drafts?.map((draft, index) => (
              <TimelineItem
                sx={{
                  "&:before": {
                    content: "unset",
                  },
                }}
              >
                <TimelineSeparator>
                  <TimelineDot />
                  <TimelineConnector />
                </TimelineSeparator>
                <TimelineContent>
                  <Typography>
                    {index + 1} - {format(draft.updatedAt, "yyyy-MM-dd")}
                  </Typography>
                  <Typography>{format(draft.updatedAt, "HH:mm:ss")}</Typography>
                </TimelineContent>
              </TimelineItem>
            ))}
          </Timeline>
        </Box>
      </SwipeableDrawer>
    </>
  );
};

ArticleEditor.defaultProps = {
  article: {
    title: "",
    content: "",
    publishedAt: new Date(),
    tags: [],
  } as Values,
};
